# Copyright (c) 2017-2020, Mikhail Podgurskiy
# All Rights Reserved.

# This work is dual-licensed under AGPL defined in file 'LICENSE' with
# LICENSE_EXCEPTION and the Commercial license defined in file 'COMM_LICENSE',
# which is part of this source code package.

import re

from django import forms, template
from django.db.models import Model
from django.urls import NoReverseMatch
from django.utils.encoding import force_str
from django.utils.html import conditional_escape

from viewflow.contrib import auth
from viewflow.forms import FormLayout
from viewflow.urls import Site, Viewset, current_viewset_reverse
from viewflow.utils import camel_case_to_title

register = template.Library()

KWARG_RE = re.compile(r"(?:(\w+)=)?(.+)")


def _parse_var_and_args(tagname, parser, bits):
    variable_name, args, kwargs = None, [], {}
    if len(bits) >= 2 and bits[-2] == "as":
        variable_name = bits[-1]
        bits = bits[:-2]

    if len(bits):
        for bit in bits:
            match = KWARG_RE.match(bit)
            if not match:
                raise template.TemplateSyntaxError(
                    "Malformed arguments for {} tag".format(tagname)
                )
            name, value = match.groups()
            if name:
                kwargs[name] = parser.compile_filter(value)
            else:
                args.append(parser.compile_filter(value))

    return variable_name, args, kwargs


def _resolve_args(context, args, kwargs):
    args = [arg.resolve(context) for arg in args]
    kwargs = {
        force_str(key, "ascii"): value.resolve(context) for key, value in kwargs.items()
    }
    return args, kwargs


class BaseViewsetURLNode(template.Node):
    """
    Base class for reversing a url to a view within a viewset
    """

    def __init__(self, parser, token):
        bits = token.split_contents()
        if len(bits) < 2:
            raise template.TemplateSyntaxError(
                "get_url takes at least two arguments, an "
                "viewset and the name of a url."
            )

        self.viewset = parser.compile_filter(bits[1])
        self.view_name = parser.compile_filter(bits[2])
        self.variable_name, self.args, self.kwargs = _parse_var_and_args(
            "get_url", parser, bits[3:]
        )

    def render(self, context):
        args, kwargs = _resolve_args(context, self.args, self.kwargs)

        viewset = self.viewset.resolve(context)
        view_name = self.view_name.resolve(context)

        if not isinstance(viewset, Viewset):
            raise template.TemplateSyntaxError(
                f"reverse '{view_name}' first argument must be a viewset instance, got '{viewset}'"
            )

        try:
            current_app = context.request.current_app
        except AttributeError:
            try:
                current_app = context.request.resolver_match.namespace
            except AttributeError:
                current_app = None

        url = ""
        try:
            url = self._reverse_url(
                viewset, view_name, args, kwargs, current_app, context
            )
        except NoReverseMatch:
            if self.variable_name is None:
                raise

        if self.variable_name:
            context[self.variable_name] = url
            return ""
        else:
            if context.autoescape:
                url = conditional_escape(url)
            return url


@register.tag("reverse")
class ViewsetURLNode(BaseViewsetURLNode):
    """
    Reverse a url to a view within viewset

    Example::
        {% reverse viewset viewname args kwargs %}
    """

    def _reverse_url(self, viewset, view_name, args, kwargs, current_app, context):
        return viewset.reverse(
            view_name, args=args, kwargs=kwargs, current_app=current_app
        )


@register.tag("current_viewset_reverse")
class CurrentViewsetURLNode(BaseViewsetURLNode):
    """
    Reverse a url to a view within viewset

    Example::
        {% current_viewset_reverse viewset viewname args kwargs %}
    """

    def _reverse_url(self, viewset, view_name, args, kwargs, current_app, context):
        return current_viewset_reverse(
            context.request, viewset, view_name, args=args, kwargs=kwargs
        )


@register.tag("render")
class FormNode(template.Node):
    """
    Render a django form using google material-components-web library.

    Example:

        {% render_form form [layout] %}
    """

    default_layout = FormLayout()

    def __init__(self, parser, token):
        bits = token.split_contents()

        layout_expr = None
        if len(bits) == 2:
            tag, form_expr = bits
        elif len(bits) == 3:
            tag, form_expr, layout_expr = bits
        else:
            raise template.TemplateSyntaxError(
                "Invalid syntax in material tag, expects only form and optional layout arguments."
            )

        self.form_expr = parser.compile_filter(form_expr)
        self.layout_expr = parser.compile_filter(layout_expr) if layout_expr else None

    def render(self, context):
        form = self.form_expr.resolve(context)
        if not isinstance(form, forms.BaseForm):
            raise template.TemplateSyntaxError(
                "material tag first argument must be a form"
            )

        layout = None
        if self.layout_expr:
            layout = self.layout_expr.resolve(context)
        if layout and not isinstance(layout, FormLayout):
            raise template.TemplateSyntaxError(
                "material tag second argument must be a layout"
            )

        return layout.render(form) if layout else self.default_layout.render(form)


@register.tag("get_absolute_url")
class AbsoluteURLNode(template.Node):
    """
    Call viewset.get_absolute_url for an object from site

    Example::
        {% get_absolute_url site view.object %}
    """

    def __init__(self, parser, token):
        bits = token.split_contents()
        if len(bits) < 3:
            raise template.TemplateSyntaxError(
                "viewset_url takes at least two arguments, a " "site and an object."
            )

        self.site = parser.compile_filter(bits[1])
        self.object = parser.compile_filter(bits[2])

        self.variable_name = None
        if len(bits) >= 2 and bits[-2] == "as":
            self.variable_name = bits[-1]

    def render(self, context):
        site = self.site.resolve(context)
        if not isinstance(site, Site):
            raise template.TemplateSyntaxError(
                "get_absolute_url first argument must be a site instance"
            )

        obj = self.object.resolve(context)
        if not isinstance(obj, Model):
            raise template.TemplateSyntaxError(
                "get_absolute_url third argument must be a model instance"
            )

        url = ""
        try:
            url = site.get_absolute_url(context.request, obj)
        except NoReverseMatch:
            if self.variable_name is None:
                raise

        if self.variable_name:
            context[self.variable_name] = url
            return ""
        else:
            if context.autoescape:
                url = conditional_escape(url)
            return url


@register.filter
def verbose_name(obj):
    """Return model verbose name."""
    if isinstance(obj, Model):
        type(obj)._meta.verbose_name
    return obj._meta.verbose_name


@register.filter
def verbose_name_plural(obj):
    """Return model verbose name in plural mode."""
    if isinstance(obj, Model):
        type(obj)._meta.verbose_name_plural
    return obj._meta.verbose_name_plural


@register.filter
def has_perm(app, user):
    """
    Check a user access rights for an application

    Example:

        {% if request.resolver_match.app|has_perm:request %}
            {% include app.menu_template_name app=request.resolver_match.app only %}
        {% endif %}
    """
    return app.has_view_permission(user)


@register.filter
def user_avatar_url(user):
    """Lookup for the user avatar."""
    return auth.get_user_avatar_url(user)


@register.filter
def list_column_order(column_def, list_view):
    """Sort order for the column from the list view."""
    if hasattr(list_view, "columns_order"):
        position, order = None, list_view.columns_order.get(column_def)
        if order:
            position = list(list_view.columns_order.keys()).index(column_def) + 1
            return {"position": position, "sort": order}
    return None


@register.filter
def list_page_data(page, list_view):
    """Formatted page data for a table.

    Returned data is a list of list of cell values zipped with column definitions.
    [[(column, value), (column, value), ...], ...]
    """
    return list_view.get_page_data(page)


@register.filter
def list_order(request_kwargs, list_view):
    return request_kwargs.get(list_view.ordering_kwarg, "")


@register.filter(name="class_title")
def class_title(obj):
    # If title attribute exists, return it
    if hasattr(obj, "title"):
        return obj.title

    # Otherwise, process class name
    class_name = obj.__class__.__name__

    # Remove "Form" and "Wizard" substrings
    cleaned_name = (
        class_name.replace("Form", "").replace("Wizard", "").replace("View", "")
    )

    if cleaned_name:
        return camel_case_to_title(cleaned_name)
    else:
        return camel_case_to_title(class_name)



# Copyright (c) 2017-2020, Mikhail Podgurskiy
# All Rights Reserved.

# This work is dual-licensed under AGPL defined in file 'LICENSE' with
# LICENSE_EXCEPTION and the Commercial licence defined in file 'COMM_LICENSE',
# which is part of this source code package.

import datetime
import decimal
import json
import uuid
from typing import List
from django.db import DEFAULT_DB_ALIAS, models, router
from django.db.models import signals
from django.db.models.sql.where import WhereNode, AND
from django.utils.duration import duration_iso_string
from django.utils.functional import Promise
from django.utils.timezone import is_aware


class CompositeKey(models.AutoField):

    class Key(dict):
        """Dictionary with json-compatible string conversion."""
        def __str__(self):
            return json.dumps(self)

        def __hash__(self):
            return hash(
                tuple(self[key] for key in sorted(self.keys()))
            )

    def __init__(self, columns: List[str], **kwargs):
        self.columns = columns
        super().__init__(primary_key=True, **kwargs)

    def contribute_to_class(self, cls, name, private_only=False):
        self.set_attributes_from_name(name)
        self.model = cls
        self.concrete = False
        self.editable = False
        self.column = self.columns[0]            # for default order_by
        cls._meta.add_field(self, private=True)  # virtual field
        cls._meta.setup_pk(self)                 # acts as pk

        if not getattr(cls, self.attname, None):
            setattr(cls, self.attname, self)

        def delete(inst, using=None, keep_parents=False):
            using = using or router.db_for_write(self.model, instance=inst)

            signals.pre_delete.send(
                sender=cls, instance=inst, using=using
            )

            query = cls._default_manager.filter(**self.__get__(inst))
            query._raw_delete(using)

            for column in self.columns:
                setattr(inst, column, None)

            signals.post_delete.send(
                sender=cls, instance=inst, using=using
            )

        cls.delete = delete

    def get_prep_value(self, value):
        return self.to_python(value)

    def to_python(self, value):
        if value is None or isinstance(value, CompositeKey.Key):
            return value
        return CompositeKey.Key(json.loads(value))

    def to_json(self, value):
        if isinstance(value, datetime.datetime):
            result = value.isoformat()
            if value.microsecond:
                result = result[:23] + result[26:]
            if result.endswith('+00:00'):
                result = result[:-6] + 'Z'
            return result
        elif isinstance(value, datetime.date):
            return value.isoformat()
        elif isinstance(value, datetime.time):
            if is_aware(value):
                raise ValueError("JSON can't represent timezone-aware times.")
            result = value.isoformat()
            if value.microsecond:
                result = result[:12]
            return result
        elif isinstance(value, datetime.timedelta):
            return duration_iso_string(value)
        elif isinstance(value, (decimal.Decimal, uuid.UUID, Promise)):
            return str(value)
        return value

    def bulk_related_objects(self, objs, using=DEFAULT_DB_ALIAS):
        return []

    def __get__(self, instance, cls=None):
        if instance is None:
            return self

        return CompositeKey.Key({
            column: self.to_json(
                self.model._meta.get_field(column).value_from_object(instance)
            )
            for column in self.columns
        })

    def __set__(self, instance, value):
        """
        I hope it's safe to ignore!
        """
        pass


@CompositeKey.register_lookup
class Exact(models.Lookup):
    lookup_name = 'exact'

    def as_sql(self, compiler, connection):
        fields = [
            self.lhs.field.model._meta.get_field(column)
            for column in self.lhs.field.columns
        ]

        lookup_classes = [
            field.get_lookup('exact')
            for field in fields
        ]

        lookups = [
            lookup_class(field.get_col(self.lhs.alias), self.rhs[column])
            for lookup_class, field, column in zip(
                lookup_classes, fields, self.lhs.field.columns
            )
        ]

        value_constraint = WhereNode()
        for lookup in lookups:
            value_constraint.add(lookup, AND)
        return value_constraint.as_sql(compiler, connection)



import json
import os
import time
from datetime import datetime, timedelta

import boto3
import jwt
import requests
import urllib3
from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from requests import Timeout, ConnectionError
from requests.auth import HTTPBasicAuth
from rest_framework import status
from rest_framework_simplejwt.tokens import RefreshToken

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
User = get_user_model()


class IamTokenService:
    URL = 'https://iam.api.cloud.yandex.net/iam/v1/tokens'

    def __init__(self):
        authorized_key = json.loads(os.getenv('IAM'))
        self._service_account_id = authorized_key['service_account_id']
        self._key_id = authorized_key['id']
        self._private_key = authorized_key['private_key']
        self._iam_token = None
        self._expires_at = None

    def get_token(self):
        if self._expires_at and self._expires_at > datetime.now():
            return self._iam_token
        else:
            self._iam_token, self._expires_at = self._request()
        return self._iam_token

    def _request(self):
        now = int(time.time())
        payload = {
            'aud': self.URL,
            'iss': self._service_account_id,
            'iat': now,
            'exp': now + 360}
        encoded_token = jwt.encode(
            payload,
            self._private_key,
            algorithm='PS256',
            headers={'kid': self._key_id})
        response = requests.post(self.URL, json={'jwt': encoded_token}).json()
        return response['iamToken'], datetime.now() + timedelta(hours=1)


class RuSmsService:
    _url = 'https://omnichannel.mts.ru/http-api/v1/messages'

    def __init__(self, sender):
        self._login, self._password = os.getenv('MTS_SMS').split('||')
        self._sender = sender

    def send(self, phonenumber, message):
        response = requests.post(self._url,
                                 json=self._get_payload(phonenumber, self._sender, message),
                                 auth=HTTPBasicAuth(self._login, self._password),
                                 timeout=(30, 30)
                                 )
        if not status.is_success(response.status_code):
            raise ValidationError(response.text)
        result = self._get_result(response.json().get("messages")[0].get("internal_id"))
        self._handle_errors(result)

    def _get_result(self, internal_id):
        try:
            response_info = requests.post(url=f"{self._url}/info",
                                          json={"int_ids": [internal_id]},
                                          auth=HTTPBasicAuth(self._login, self._password), )
            return response_info.json().get('events_info')[0].get('events_info')[0]
        except (IndexError, TypeError, IOError):
            pass

    @staticmethod
    def _get_payload(phonenumber, sender, message):
        msisdn = f'{phonenumber.country_code}{phonenumber.national_number}'
        payload = {
            "messages": [
                {
                    "content": {
                        "short_text": message
                    },
                    "from": {
                        "sms_address": sender
                    },
                    "to": [
                        {
                            "msisdn": msisdn
                        }
                    ]
                }
            ]
        }
        return payload

    @staticmethod
    def _handle_errors(result):
        if not result:
            return
        internal_errors = result.get('internal_errors')
        if not internal_errors:
            return result
        elif 10802 in internal_errors:
            raise ValidationError('phonenumber_format_error')
        elif 10803 in internal_errors:
            raise ValidationError('unsupported_code_country')
        elif 10904 in internal_errors:
            raise ValidationError('phonenumber_blacklist')
        elif 10100 in internal_errors:
            raise ValidationError('incorrect_sender')
        else:
            raise ValidationError(internal_errors)


class YandexObjectService:
    _bucket = 'expressmoney'
    _filename = None

    def __init__(self):
        self._s3 = boto3.session.Session().client(service_name='s3', endpoint_url='https://storage.yandexcloud.net')

    def generate_key(self, filename):
        self._filename = filename
        prepared_form_fields = self._s3.generate_presigned_post(Bucket=self._bucket,
                                                                Key=f'files/{self._filename}',
                                                                Conditions=[["starts-with", "$key", "files"], ],
                                                                ExpiresIn=60 * 60)
        return json.dumps(prepared_form_fields)


class ScoringService:
    _url = 'https://api.expressmoney.com/ru_scoring/pdl_ru/scoring'

    def __init__(self):
        self._headers = {'Authorization': f'Token {os.getenv("SERVICE_TOKEN")}'}

    def request(self, data: dict) -> dict:
        attempts = 0
        response = self._request(attempts, data)
        order_id = data.get("order_id")
        if status.is_success(response.status_code):
            return response.json()
        elif response.status_code == status.HTTP_400_BAD_REQUEST and order_id:
            response = requests.get(f'{self._url}/{order_id}', headers=self._headers)
            if status.is_success(response.status_code):
                return response.json()
        raise ValidationError(response.text[:1024])

    def _request(self, attempts, data):
        try:
            response = requests.post(self._url, headers=self._headers, json=data, verify=False, timeout=(300, 300))
        except ConnectionError as e:
            attempts += 1
            if attempts <= 3:
                time.sleep(1)
                response = self._request(attempts, data)
            else:
                raise e
        return response


class RuPartnerService:
    SERVICE_URL = 'https://bbaic3serh6adpvivht0.containers.yandexcloud.net'

    def __init__(self):
        self._iam_token_service = IamTokenService()

    def create_lead(self, referral_id: int):
        url = f'{self.SERVICE_URL}/ru_partners/leads/lead'
        requests.post(url, headers=self._get_header(referral_id), data={'description': 1})

    def create_referral(self, referral_id, partner_code):
        url = f'{self.SERVICE_URL}/ru_partners/referrals/referral'
        requests.post(url, headers=self._get_header(referral_id), data={'partner_code': partner_code})

    def send_order_postback(self, user_id: int, order_id: int, utm_source: str, transaction_id: str, amount):
        url = f'{self.SERVICE_URL}/ru_partners/cpa/order_client_post_back'
        requests.post(url,
                      headers=self._get_header(user_id),
                      data={
                          'order_id': order_id,
                          'utm_source': utm_source,
                          'transaction_id': transaction_id,
                          'amount': amount
                      })

    def send_approved_postback(self, user_id: int, order_id: int, utm_source: str, transaction_id: str, amount):
        url = f'{self.SERVICE_URL}/ru_partners/cpa/approved_postback'
        requests.post(url,
                      headers=self._get_header(user_id),
                      data={
                          'order_id': order_id,
                          'utm_source': utm_source,
                          'transaction_id': transaction_id,
                          'amount': amount
                      })

    def send_loan_postback(self, user_id: int, loan_id: int, utm_source: str, transaction_id: str, amount):
        url = f'{self.SERVICE_URL}/ru_partners/cpa/loan_client_post_back'
        requests.post(url, headers=self._get_header(user_id), data={
            'loan_id': loan_id,
            'utm_source': utm_source,
            'transaction_id': transaction_id,
            'amount': amount
        })

    def send_add_profile_postback(self, user_id: int, utm_source: str):
        url = f'{self.SERVICE_URL}/ru_partners/cpa/send_add_profile_postback'
        requests.post(url, headers=self._get_header(user_id), data={
            'user_id': user_id,
            'utm_source': utm_source,
        })

    def send_order_postback_deeplink(self, user_id: int, order_id: int, phone_number: str, amount):
        url = f'{self.SERVICE_URL}/ru_partners/cpa/order_user_post_back'
        requests.post(url,
                      headers=self._get_header(user_id),
                      data={
                          'order_id': order_id,
                          'phone_number': phone_number,
                          'amount': amount
                      })

    def send_loan_postback_deeplink(self, user_id: int, loan_id: int, phone_number: str, amount):
        url = f'{self.SERVICE_URL}/ru_partners/cpa/loan_user_post_back'
        requests.post(url,
                      headers=self._get_header(user_id),
                      data={
                          'loan_id': loan_id,
                          'phone_number': phone_number,
                          'amount': amount
                      })

    def get_cpa_client(self, user_id: int):
        url = f'{self.SERVICE_URL}/ru_partners/cpa/cpa_client/{user_id}'
        return requests.get(url, headers=self._get_header(user_id))

    def get_cpa_user(self, user_id: int, phone_number: str):
        url = f'{self.SERVICE_URL}/ru_partners/cpa/cpa_user/{phone_number}'
        return requests.get(url, headers=self._get_header(user_id))

    def set_user_id(self, user_id: int, cpaclient: int):
        url = f'{self.SERVICE_URL}/ru_partners/cpa/set_user_id'
        return requests.post(url, headers=self._get_header(user_id), json={'user_id': user_id, 'cpaclient': cpaclient})

    def is_full_deal(self, user_id: int):
        url = f'{self.SERVICE_URL}/ru_partners/cpa/full_deal/is_full_deal'
        return requests.post(url, headers=self._get_header(user_id))

    def _get_header(self, user_id):
        refresh = RefreshToken.for_user(User.objects.get(pk=user_id))
        headers = {
            'Authorization': f'Bearer {self._iam_token_service.get_token()}',
            'Auth': f'Bearer {refresh.access_token}'
        }
        return headers


class AuthService:
    SERVICE_URL = 'https://bba6ogflnltlachiiqiu.containers.yandexcloud.net'

    def __init__(self):
        self.attempts = 0
        self._iam_token_service = IamTokenService()

    def list_sms(self, user_id: int):
        url = f'{self.SERVICE_URL}/auth/sms/sms'
        return requests.get(url, headers=self._get_header(user_id), timeout=(120, 120))

    def get_code(self, user_id: int):
        """Код для формирования url для автоматического входа без пароля"""
        try:
            headers = self._get_header(user_id)
            url = f'{self.SERVICE_URL}/auth/auto_login/code/{user_id}'
            response = requests.get(url, headers=headers, timeout=(120, 120))
            if status.is_success(response.status_code):
                return response.json().get('code')
            elif response.status_code == status.HTTP_404_NOT_FOUND:
                url = f'{self.SERVICE_URL}/auth/auto_login/code'
                response = requests.post(url, headers=headers, timeout=(120, 120))
                if not status.is_success(response.status_code):
                    raise ValidationError(response.text)
                return response.json().get('code')
            else:
                raise ValidationError(response.text)
        except Timeout as e:
            if self.attempts < 5:
                self.attempts += 1
                return self.get_code(user_id)
            else:
                raise e

    def create_user_ext(self, user_data: dict):
        user_id = int(os.getenv('SERVICE_USER_ID'))
        url = f'{self.SERVICE_URL}/auth/ext/ext'
        return requests.post(url, headers=self._get_header(user_id), json=user_data, timeout=(120, 120))

    def update_last_login(self, user_id):
        url = f'{self.SERVICE_URL}/auth/ext/user/{user_id}/update_last_login'
        return requests.post(url, headers=self._get_header(user_id))

    def get_user_id(self, phonenumber: str):
        url = f'{self.SERVICE_URL}/auth/ext/user/get_user_id'
        response = requests.post(url, data={'phonenumber': phonenumber}, headers=self._get_header_allowany())
        return response.json().get('user_id')

    def list_provider(self, user_id):
        url = f'{self.SERVICE_URL}/auth/providers/provider'
        return requests.get(url, headers=self._get_header(user_id))

    def get_provider_data(self, provider_token, provider):
        url = f'{self.SERVICE_URL}/auth/providers/provider_data'
        return requests.post(url, data={'provider_token': provider_token,
                                        'provider': provider}, headers=self._get_header_allowany())

    def get_provider_data_id(self, provider_id):
        url = f'{self.SERVICE_URL}/auth/providers/provider_data/{provider_id}'
        return requests.get(url, headers=self._get_header_allowany())

    def _get_header_allowany(self) -> dict:
        headers = {
            'Authorization': f'Bearer {self._iam_token_service.get_token()}'
        }
        return headers

    def _get_header(self, user_id: int) -> dict:
        refresh = RefreshToken.for_user(User.objects.get(pk=user_id))
        headers = {
            'Authorization': f'Bearer {self._iam_token_service.get_token()}',
            'Auth': f'Bearer {refresh.access_token}'
        }
        return headers


class RuProfilesService:
    SERVICE_URL = 'https://bbaqp3g26tvqqijq9rdn.containers.yandexcloud.net'

    def __init__(self, api_version=''):
        self._iam_token_service = IamTokenService()
        self._api_version = api_version

    def create_profile(self, user_id: int, profile_data: dict):
        url = f'{self.SERVICE_URL}/ru_profiles/profiles/profile'
        return requests.post(url, headers=self._get_header(user_id), json=profile_data, timeout=(120, 120))

    def request_inn(self, payload):
        user_id = int(os.getenv('SERVICE_USER_ID'))
        url = f'{self.SERVICE_URL}/ru_profiles/profiles/inn/request_inn'
        return requests.post(url, headers=self._get_header(user_id), json=payload, timeout=(120, 120))

    def request_prohibition(self, payload):
        user_id = int(os.getenv('SERVICE_USER_ID'))
        url = f'{self.SERVICE_URL}/ru_profiles/profiles/prohibition/request_prohibition'
        return requests.post(url, headers=self._get_header(user_id), json=payload, timeout=(120, 120))

    def set_inn(self, user_id, payload):
        url = f'{self.SERVICE_URL}/ru_profiles/profiles/profile/{user_id}/set_inn'
        return requests.post(url, headers=self._get_header(user_id), json=payload, timeout=(120, 120))

    def _get_header(self, user_id: int) -> dict:
        refresh = RefreshToken.for_user(User.objects.get(pk=user_id))
        headers = {
            'Authorization': f'Bearer {self._iam_token_service.get_token()}',
            'Auth': f'Bearer {refresh.access_token}',
            'Accept': f'application/json; version={self._api_version}'
        }
        return headers


class RuLoansService:
    SERVICE_URL = 'https://bbavf0h1b2vaed78vvsf.containers.yandexcloud.net'

    def __init__(self):
        self._iam_token_service = IamTokenService()

    def create_order(self, user_id: int, order_data: dict):
        url = f'{self.SERVICE_URL}/ru_loans/orders/order'
        return requests.post(url, headers=self._get_header(user_id), json=order_data, timeout=(120, 120))

    def create_loan(self, user_id: int, loan_data: dict):
        url = f'{self.SERVICE_URL}/ru_loans/loans/loan'
        return requests.post(url, headers=self._get_header(user_id), json=loan_data, timeout=(120, 120))

    def body_issue(self, user_id, loan_id, bank_card_id):
        payload = {
            'loan': loan_id,
            'bank_card_id': bank_card_id
        }
        url = f'{self.SERVICE_URL}/ru_loans/operations/body_issue'
        return requests.post(url, headers=self._get_header(user_id), json=payload, timeout=(120, 120))

    def create_provider_bonus(self, user_id):
        url = f'{self.SERVICE_URL}/ru_loans/bonuses/bonus/create_provider'
        return requests.post(url, headers=self._get_header(user_id), timeout=(120, 120))

    def send_sign(self, user_id: int, order_id: int):
        url = f'{self.SERVICE_URL}/ru_loans/orders/order/{order_id}/send_sign'
        return requests.post(url, headers=self._get_header(user_id), timeout=(120, 120))

    def _get_header(self, user_id: int) -> dict:
        refresh = RefreshToken.for_user(User.objects.get(pk=user_id))
        headers = {
            'Authorization': f'Bearer {self._iam_token_service.get_token()}',
            'Auth': f'Bearer {refresh.access_token}',
        }
        return headers


class FraudProcessService:
    SERVICE_URL = 'https://bbaqp3g26tvqqijq9rdn.containers.yandexcloud.net'

    def __init__(self, api_version=''):
        self._iam_token_service = IamTokenService()
        self._api_version = api_version

    def create(self, user_id, comment):
        url = f'{self.SERVICE_URL}/ru_profiles/profiles/fraud_process'
        response = requests.post(url, headers=self._get_header(user_id), json={'comment': comment})
        return response

    def _get_header(self, user_id: int) -> dict:
        refresh = RefreshToken.for_user(User.objects.get(pk=user_id))
        headers = {
            'Authorization': f'Bearer {self._iam_token_service.get_token()}',
            'Auth': f'Bearer {refresh.access_token}',
            'Accept': f'application/json; version={self._api_version}'
        }
        return headers


class HolderNameProcessService:
    SERVICE_URL = 'https://bbaqp3g26tvqqijq9rdn.containers.yandexcloud.net'

    def __init__(self, api_version=''):
        self._iam_token_service = IamTokenService()
        self._api_version = api_version

    def create(self, user_id, bank_card_id):
        url = f'{self.SERVICE_URL}/ru_profiles/profiles/holder_name_process'
        response = requests.post(url, headers=self._get_header(user_id), json={'bank_card_id': bank_card_id})
        return response

    def _get_header(self, user_id: int) -> dict:
        refresh = RefreshToken.for_user(User.objects.get(pk=user_id))
        headers = {
            'Authorization': f'Bearer {self._iam_token_service.get_token()}',
            'Auth': f'Bearer {refresh.access_token}',
            'Accept': f'application/json; version={self._api_version}'
        }
        return headers


class BlackListService:
    SERVICE_URL = 'https://bba6ogflnltlachiiqiu.containers.yandexcloud.net'

    def __init__(self):
        self._iam_token_service = IamTokenService()

    def create(self, user_id, cause, comment):
        url = f'{self.SERVICE_URL}/auth/black_list/black_list'
        response = requests.post(url, headers=self._get_header(user_id), json={'cause': cause, 'comment': comment})
        return status.is_success(response.status_code)

    def retrieve(self, user_id):
        url = f'{self.SERVICE_URL}/auth/black_list/black_list/{user_id}'
        response = requests.get(url, headers=self._get_header(user_id))
        return response.json().get('cause'), response.json().get('comment')

    def delete(self, user_id):
        url = f'{self.SERVICE_URL}/auth/black_list/black_list/{user_id}'
        response = requests.delete(url, headers=self._get_header(user_id))
        return status.is_success(response.status_code)

    def find(self, user_id):
        url = f'{self.SERVICE_URL}/auth/black_list/black_list/find'
        response = requests.post(url, headers=self._get_header(user_id))
        return response.json().get('is_black_list'), response.json().get('cause')

    def _get_header(self, user_id: int) -> dict:
        refresh = RefreshToken.for_user(User.objects.get(pk=user_id))
        headers = {
            'Authorization': f'Bearer {self._iam_token_service.get_token()}',
            'Auth': f'Bearer {refresh.access_token}',
            'Accept': f'application/json'
        }
        return headers



from django.db import models


class Profile(models.Model):
    BLANK = ''
    MATCH = 'MATCH'
    NOT_MATCH = 'NOT_MATCH'
    IN_PROCESS = 'IN_PROCESS'
    NOT_DISPLAYED = 'NOT_DISPLAYED'
    HOLDER_NAME_STATUS = {
        BLANK: '',
        MATCH: 'Имя совпадает',
        NOT_MATCH: 'Имя НЕ совпадает',
        IN_PROCESS: 'Идет проверка',
        NOT_DISPLAYED: 'Имя НЕ ОТОБРАЖАЕТСЯ',
    }
    user_id = models.PositiveIntegerField(primary_key=True)
    created = models.DateTimeField('Создано', auto_now_add=True)
    updated = models.DateTimeField('Изменено', auto_now=True)
    # ФИО
    first_name = models.CharField('Имя', max_length=32)
    last_name = models.CharField('Фамилия', max_length=32)
    middle_name = models.CharField('Отчество', max_length=32)
    birth_date = models.DateField('Дата рождения')
    # Идентификация, верификация
    identified = models.DateTimeField('Идентифицирован', null=True, blank=True)
    verified = models.DateTimeField('Верифицирован', null=True, blank=True)
    holder_name_status = models.CharField('Имя на карте', max_length=32,
                                          choices=sorted(HOLDER_NAME_STATUS.items()), blank=True)
    # Паспорт
    passport_serial = models.CharField('Серия', max_length=4)
    passport_number = models.CharField('Номер', max_length=6)
    passport_birth_place = models.CharField('Место рождения', max_length=256, blank=True)
    passport_issue_name = models.CharField('Кем выдан', max_length=256, blank=True)
    passport_code = models.CharField('Код паспорта', max_length=16)
    passport_date = models.DateField('Дата выдачи', )
    # Адрес
    postal_code = models.CharField('Индекс', max_length=16, blank=True)
    state = models.CharField('Регион', max_length=256, blank=True)
    city = models.CharField('Город', max_length=256, blank=True)
    street = models.CharField('Улица', max_length=256, blank=True)
    house = models.CharField('Дом', max_length=8, blank=True)
    apt = models.CharField('Кв', max_length=8, blank=True)
    # Прочее
    snils = models.CharField('Снилс', max_length=64, null=True, blank=True)
    comment = models.TextField('Комментарий', max_length=5120, blank=True)
    inn = models.CharField('ИНН', max_length=12, null=True, blank=True)

    def __str__(self):
        return f'{self.user_id}'

    class Meta:
        managed = False
        verbose_name = 'Профиль'
        verbose_name_plural = 'Профили'
        unique_together = (('passport_serial', 'passport_number', ),
                           ('first_name', 'last_name', 'middle_name', 'birth_date'),
                           )
        constraints = (models.UniqueConstraint(fields=('first_name', 'last_name', 'middle_name', 'birth_date'),
                                               name='unique_full_name'),
                       models.UniqueConstraint(fields=('passport_serial', 'passport_number', ),
                                               name='unique_passport'),
                       )


class Lost(models.Model):
    profile = models.OneToOneField('Profile', on_delete=models.CASCADE, primary_key=True, verbose_name='Профиль')
    loan_closed = models.DateField('Займ закрыт', help_text='Дата закрытия последнего займа')
    score = models.DecimalField('Скор', max_digits=4, decimal_places=3, help_text='Рассчитанный  нами')
    score_date = models.DateField('Дата скора', help_text='Дата расчета скора и заполнения данных по займам')
    amount_approved = models.DecimalField('Одобренная сумма', max_digits=7, decimal_places=0)
    nbki_score = models.SmallIntegerField('Скор НБКИ')
    total_loans = models.IntegerField('Всего займов', help_text='Всего займов')
    first_loan = models.IntegerField('первый заём', help_text='Дней с первого займа')
    last_loan = models.IntegerField('Последний заём', help_text='Дней с последнего займа')
    interests_sum = models.IntegerField('Выплачено процентов')
    body_sum = models.IntegerField('Сумма займов', help_text='Сумма тела всех займов')
    body_max = models.IntegerField('Макс заём', help_text='Максимальная сумма займа')

    def save(self, *args, **kwargs):
        pass

    def __str__(self):
        return f'{self.pk}'

    class Meta:
        managed = False
        verbose_name = 'Потерянный клиент'
        verbose_name_plural = 'Потерянные клиенты'



# Copyright (c) 2017-2020, Mikhail Podgurskiy
# All Rights Reserved.

# This work is licensed under the Commercial license defined in file
# 'COMM_LICENSE', which is part of this source code package.

from django.core.exceptions import PermissionDenied, SuspiciousOperation
from django.forms import modelform_factory
from django.http import HttpResponseRedirect
from django.urls import path
from django.utils.translation import gettext_lazy as _
from viewflow.urls import ViewsetMeta
from viewflow.views import Action, UpdateModelView
from viewflow.utils import viewprop, get_object_data

from viewflow.forms import ModelForm


from .base import State, TransitionBoundMethod


class ModelTransitionView(UpdateModelView):
    template_name_suffix = "_transition"

    def get_transition_fields(self, request, obj, slug):
        return self.viewset.get_transition_fields(request, obj, slug)

    def get_object_data(self):
        """List of object fields to display.
        Choice fields values are expanded to readable choice label.
        """
        return get_object_data(self.object)

    def get_template_names(self):
        """
        List of templates for the view.
        If no `self.template_name` defined, uses::
             [<app_label>/<model_label>_<suffix>.html,
              <app_label>/<model_label>_transition.html,
              'viewflow/views/transition.html']
        """
        if self.template_name is None:
            opts = self.model._meta
            return [
                "{}/{}{}.html".format(
                    opts.app_label, opts.model_name, self.template_name_suffix
                ),
                "{}/{}_transition.html".format(opts.app_label, opts.model_name),
                "viewflow/views/transition.html",
            ]
        return [self.template_name]

    def get_form_class(self):
        if self.form_class is None:
            return modelform_factory(
                self.model,
                form=ModelForm,
                fields=self.get_transition_fields(
                    self.request, self.object, self.kwargs["slug"]
                ),
            )
        else:
            return super().get_form_class()

    def form_valid(self, form):
        self.object = form.save()
        self.transition()
        return HttpResponseRedirect(self.get_success_url())

    def get_object(self, *args, **kwargs):
        obj = super().get_object(*args, **kwargs)

        flow = self.viewset.get_object_flow(self.request, obj)
        slug = self.kwargs["slug"]
        self.transition = (
            getattr(flow, slug, None) if not slug.startswith("_") else None
        )
        if not self.transition or not isinstance(
            self.transition, TransitionBoundMethod
        ):
            raise SuspiciousOperation

        if not self.transition.has_perm(self.request.user):
            raise PermissionDenied

        if not self.transition.can_proceed():
            raise PermissionDenied(_("Transition is not allowed"))

        return obj


class FlowViewsMixin(metaclass=ViewsetMeta):
    flow_state = None
    transition_view_class = ModelTransitionView

    def get_flow_state(self, request) -> State:
        if self.flow_state is None:
            raise ValueError("flow_state attribute is not defined.")
        return self.flow_state

    def get_object_flow(self, request, obj):
        try:
            return self.get_flow_state()._owner(obj)
        except TypeError:
            raise ValueError(
                "%s has no constructor with single argument. Please "
                "redefine .get_object_flow(self, request, obj) on the "
                "viewset" % self.flow_state._owner
            )

    def get_transition_fields(self, request, obj, slug):
        return []

    def get_detail_page_object_actions(self, request, obj, *actions):
        state = self.get_flow_state(request)
        flow = self.get_object_flow(request, obj)
        transitions = [
            (
                transition,
                transition.conditions_met(flow),
                transition.has_perm(flow, request.user),
            )
            for transition in state.get_outgoing_transitions(state.get(flow))
        ]

        additional_actions = [
            Action(
                name=transition.label,
                url=self.reverse("transition", args=[obj.pk, transition.slug]),
            )
            for transition, conditions_met, has_perm in transitions
            if conditions_met and has_perm
        ]

        if actions:
            additional_actions = (*additional_actions, *actions)

        return super().get_detail_page_object_actions(request, obj, *additional_actions)

    """
    Default transition view
    """

    def get_transition_view_kwargs(self, **kwargs):
        view_kwargs = {**self.transition_view_kwargs, **kwargs}
        return self.filter_kwargs(self.transition_view_class, **view_kwargs)

    @viewprop
    def transition_view_kwargs(self):
        return {}

    @viewprop
    def transition_view(self):
        return self.transition_view_class.as_view(**self.get_transition_view_kwargs())

    def _get_urls(self):
        urlpatterns = super()._get_urls()
        urlpatterns.append(
            path(
                "<path:pk>/transition/<slug:slug>/",
                self.transition_view,
                name="transition",
            )
        )
        return urlpatterns



from django.utils.timezone import now

from viewflow import this
from viewflow.utils import is_owner
from ..base import Node
from ..activation import Activation, leading_tasks_canceled, has_manage_permission
from ..status import STATUS, PROCESS
from ..signals import task_started, task_finished, flow_started
from . import mixins


class StartActivation(mixins.NextNodeActivationMixin, Activation):
    """Task activation that creates new process instance."""

    @classmethod
    def create(cls, flow_task, prev_activation, token):
        flow_class = flow_task.flow_class

        process = flow_class.process_class(flow_class=flow_class)
        task = flow_class.task_class(
            flow_task=flow_task, process=process, started=now()
        )

        return cls(task)

    @Activation.status.transition(source=STATUS.NEW)
    def execute(self):
        task_started.send(sender=self.flow_class, process=self.process, task=self.task)
        self.process.save()
        with self.flow_class.lock(self.process.pk):
            self.complete()
            flow_started.send(
                sender=self.flow_class,
                process=self.process,
                task=self.task,
            )
            self.activate_next()


class StartHandleActivation(StartActivation):
    @Activation.status.transition(
        source=STATUS.DONE,
        target=STATUS.CANCELED,
        conditions=[leading_tasks_canceled],
        permission=has_manage_permission,
    )
    def undo(self):
        # undo
        if self.flow_task._undo_func is not None:
            self.flow_task._undo_func(self)

        # cancel process
        self.process.finished = now()
        self.process.status = PROCESS.CANCELED
        self.process.save()

        # finish task
        self.task.finished = now()
        self.task.save()


class StartViewActivation(StartActivation):
    """Task activation that creates new process instance."""

    @Activation.status.transition(source=STATUS.NEW, target=STATUS.STARTED)
    def start(self, request):
        self.task.owner = request.user

    @Activation.status.transition(source=STATUS.STARTED, target=STATUS.DONE)
    def complete(self):
        super().complete.original()

    @Activation.status.transition(source=STATUS.STARTED)
    def execute(self):
        super().execute.original()

    @Activation.status.super()
    def undo(self):
        # undo
        if self.flow_task._undo_func is not None:
            self.flow_task._undo_func(self)

        # cancel process
        self.process.finished = now()
        self.process.status = PROCESS.CANCELED
        self.process.save()

        super().undo.original()


class Start(
    mixins.NextNodeMixin,
    mixins.NodePermissionMixin,
    Node,
):
    """Start a flow from django view."""

    activation_class = StartViewActivation

    task_type = "HUMAN_START"

    shape = {
        "width": 50,
        "height": 50,
        "svg": """
            <circle class="event" cx="25" cy="25" r="25"/>
        """,
    }

    bpmn_element = "startEvent"

    def __init__(self, view, undo_func=None, **kwargs):
        super().__init__()
        self._task_result_summary = "Process launched"
        self._start_view = view
        self._undo_func = undo_func

    def can_execute(self, user, task=None):
        """
        Check whether the user is authorized to start a flow.

        This method checks whether the user is authorized to start a flow based on the
        owner and owner permission attributes.
        """
        if not user.is_authenticated:
            return False

        if task and task.status != STATUS.NEW:
            return False

        from django.contrib.auth import get_user_model

        if self._owner:
            if callable(self._owner):
                return self._owner(user)
            else:
                owner = get_user_model()._default_manager.get(**self._owner)
                return is_owner(owner, user)

        elif self._owner_permission:
            obj = None
            if self._owner_permission_obj:
                if callable(self._owner_permission_obj):
                    obj = self._owner_permission_obj()
                else:
                    obj = self._owner_permission_obj

            return user.has_perm(self._owner_permission, obj=obj) or user.has_perm(
                self._owner_permission
            )

        else:
            """
            No restriction
            """
            return True


class StartHandle(mixins.NextNodeMixin, Node):
    """Start flow from code."""

    task_type = "START"

    shape = {
        "width": 50,
        "height": 50,
        "svg": """
            <circle class="event" cx="25" cy="25" r="25"/>
        """,
    }

    bpmn_element = "startEvent"

    activation_class = StartHandleActivation

    def __init__(self, func=None, undo_func=None):
        super().__init__()
        self._func = func
        self._undo_func = undo_func

    def _create_wrapper_function(self, origin_func):
        def func(**kwargs):
            activation = self.activation_class.create(self, None, None)
            result = (
                origin_func(activation, **kwargs) if origin_func else activation.process
            )
            activation.execute()
            return result

        return func

    def run(self, **kwargs):
        func = this.resolve(self.flow_class.instance, self._func)
        wrapper = self._create_wrapper_function(func)
        return wrapper(**kwargs)



from django.core.exceptions import FieldDoesNotExist
from django.db.models import Q
from django.db.models.constants import LOOKUP_SEP
from django.utils.text import smart_split, unescape_string_literal


def lookup_spawns_duplicates(opts, lookup_path):
    """
    from django.contrib.admin.utils import lookup_spawns_duplicates

    Return True if the given lookup path spawns duplicates.
    """
    lookup_fields = lookup_path.split(LOOKUP_SEP)
    # Go through the fields (following all relations) and look for an m2m.
    for field_name in lookup_fields:
        if field_name == 'pk':
            field_name = opts.pk.name
        try:
            field = opts.get_field(field_name)
        except FieldDoesNotExist:
            # Ignore query lookups.
            continue
        else:
            if hasattr(field, 'path_infos'):
                # This field is a relation; update opts to follow the relation.
                path_info = field.path_infos
                opts = path_info[-1].to_opts
                if any(path.m2m for path in path_info):
                    # This field is a m2m relation so duplicates must be
                    # handled.
                    return True
    return False


def construct_search(opts, field_name):
    """
    Implementation similar to django.contrib.admin.options.ModelAdmin.get_search_results
    """
    if field_name.startswith("^"):
        return "%s__istartswith" % field_name[1:]
    elif field_name.startswith("="):
        return "%s__iexact" % field_name[1:]
    elif field_name.startswith("@"):
        return "%s__search" % field_name[1:]

    # Use field_name if it includes a lookup.
    lookup_fields = field_name.split(LOOKUP_SEP)

    # Go through the fields, following all relations.
    prev_field = None
    for path_part in lookup_fields:
        if path_part == "pk":
            path_part = opts.pk.name
        try:
            field = opts.get_field(path_part)
        except FieldDoesNotExist:
            # Use valid query lookups.
            if prev_field and prev_field.get_lookup(path_part):
                return field_name
        else:
            prev_field = field
            if hasattr(field, "path_infos"):
                # Update opts to follow the relation.
                opts = field.path_infos[-1].to_opts

    # Otherwise, use the field with icontains.
    return "%s__icontains" % field_name


def get_search_results(queryset, search_fields, search_term):
    """
    Implementation similar to django.contrib.admin.options.ModelAdmin.get_search_results
    """
    opts = queryset.model._meta

    orm_lookups = [
        construct_search(opts, str(search_field)) for search_field in search_fields
    ]

    for bit in smart_split(search_term):
        if bit.startswith(('"', "'")) and bit[0] == bit[-1]:
            bit = unescape_string_literal(bit)
        or_queries = Q(
            *((orm_lookup, bit) for orm_lookup in orm_lookups),
            _connector=Q.OR,
        )
        queryset = queryset.filter(or_queries)

    may_have_duplicates = any(
        lookup_spawns_duplicates(opts, search_spec)
        for search_spec in orm_lookups
    )

    if may_have_duplicates:
        queryset = queryset.distinct()

    return queryset


class SearchableViewMixin(object):
    """
    The mixin for LitView to enable search capabilities
    """
    search_fields = None

    def search_enabled(self):
        return self.search_fields is not None

    def get_search_term(self):
        return self.request.GET.get('_search')

    def get_queryset(self):
        queryset = super().get_queryset()
        search_term = self.get_search_term()
        if self.search_enabled is not None and search_term is not None:
            queryset = get_search_results(queryset, self.search_fields, search_term)
        return queryset



# Copyright (c) 2017-2020, Mikhail Podgurskiy
# All Rights Reserved.

# This work is dual-licensed under AGPL defined in file 'LICENSE' with
# LICENSE_EXCEPTION and the Commercial license defined in file 'COMM_LICENSE',
# which is part of this source code package.

import datetime
import decimal
from functools import lru_cache

from django.contrib.auth.decorators import login_required
from django.core.exceptions import FieldDoesNotExist, PermissionDenied
from django.db import models
from django.forms.utils import pretty_name
from django.utils import formats, timezone
from django.utils.decorators import method_decorator
from django.utils.encoding import force_str
from django.utils.functional import cached_property
from django.utils.html import format_html
from django.views import generic

from viewflow.utils import Icon, has_object_perm, viewprop
from .filters import FilterableViewMixin
from .search import SearchableViewMixin


def _get_method_attr(data_source, method_name, attr_name, default=None):
    attr = getattr(data_source, method_name)
    try:
        return getattr(attr, attr_name)
    except AttributeError:
        if isinstance(attr, property) and hasattr(attr, "fget"):
            return getattr(attr.fget, attr_name, default)
    return default


class BaseColumn(object):
    def __init__(self, attr_name):
        self.attr_name = attr_name

    def get_value(self, obj):
        raise NotImplementedError("subclasses must implement this method.")

    def header(self):
        raise NotImplementedError("subclasses must implement this method")

    def column_type(self):
        raise NotImplementedError("subclasses must implement this method")

    def orderby(self):
        raise NotImplementedError("subclasses must implement this method")

    def format_value(self, obj, value):
        if value is None:
            return ""
        elif isinstance(value, datetime.datetime):
            return formats.localize(timezone.template_localtime(value))
        elif isinstance(value, (datetime.date, datetime.time)):
            return formats.localize(value)
        elif isinstance(value, (int, float, decimal.Decimal)):
            return formats.number_format(value)
        elif isinstance(value, (list, tuple)):
            return ", ".join(force_str(v) for v in value)
        else:
            return force_str(value)


class ModelFieldColumn(BaseColumn):
    """
    Retrieve a field value from a model.

    Field verbose name would be use as a label.
    """

    NUMBER_FIELD_TYPES = (
        models.IntegerField,
        models.DecimalField,
        models.FloatField,
    )

    BOOLEAN_FIELD_TYPES = (models.BooleanField, models.NullBooleanField)

    def __init__(self, model_field):
        super().__init__(model_field.name)
        self.model_field = model_field

    def get_value(self, obj):
        return getattr(obj, self.model_field.name)

    def header(self):
        try:
            return self.model_field.verbose_name.capitalize()
        except AttributeError:
            # field is likely a ForeignObjectRel
            return self.model_field.related_model._meta.verbose_name.capitalize()

    def column_type(self):
        if isinstance(self.model_field, ModelFieldColumn.NUMBER_FIELD_TYPES):
            return "numeric"
        elif isinstance(self.model_field, ModelFieldColumn.BOOLEAN_FIELD_TYPES):
            return "boolean"
        return "text"

    def orderby(self):
        return self.attr_name

    def format_value(self, obj, value):
        if getattr(self.model_field, "flatchoices", None):
            return dict(self.model_field.flatchoices).get(value, "")
        elif isinstance(self.model_field, ModelFieldColumn.BOOLEAN_FIELD_TYPES):
            if value is None:
                return Icon("indeterminate_check_box")
            elif value is True:
                return Icon("check_box")
            else:
                return Icon("check_box_outline_blank")
        else:
            return super().format_value(obj, value)


class DataSourceColumn(BaseColumn):
    """
    Retrieve attribute value from external data source.

    Data source attribute could be a property or callable.
    For a callable, to get the value it would be called with model
    instance.
    """

    def __init__(self, data_source, attr_name, verbose_name=None):
        super().__init__(attr_name)
        self.verbose_name = verbose_name
        self.data_source = data_source

    def _get_attr_boolean(self):
        return _get_method_attr(self.data_source, self.attr_name, "boolean", False)

    def _get_attr_empty_value(self):
        return _get_method_attr(self.data_source, self.attr_name, "empty_value")

    def get_value(self, obj):
        attr = getattr(self.data_source, self.attr_name)
        if callable(attr):
            attr = attr(obj)
        if attr is None:
            attr = self._get_attr_empty_value()
        return attr

    def header(self):
        if self.verbose_name is not None:
            return self.verbose_name
        attr = getattr(self.data_source, self.attr_name)
        if hasattr(attr, "short_description"):
            return attr.short_description
        elif isinstance(attr, property) and hasattr(attr, "fget"):
            if hasattr(attr.fget, "short_description"):
                return attr.fget.short_description
            else:
                return pretty_name(self.attr_name)
        elif callable(attr):
            return "--" if attr.__name__ == "<lambda>" else pretty_name(attr.__name__)
        else:
            return pretty_name(self.attr_name)

    def column_type(self):
        is_boolean = _get_method_attr(self.data_source, self.attr_name, "boolean", None)
        if is_boolean:
            return "boolean"
        return _get_method_attr(self.data_source, self.attr_name, "column_type", "text")

    def orderby(self):
        return _get_method_attr(
            self.data_source, self.attr_name, "orderby_column", None
        )

    def format_value(self, obj, value):
        if self._get_attr_boolean():
            if value is None:
                return Icon("indeterminate_check_box")
            elif value is True:
                return Icon("check_box")
            else:
                return Icon("check_box_outline_blank")
        else:
            return super().format_value(obj, value)


class ObjectAttrColumn(DataSourceColumn):
    """
    Retrieve attribute value from a model instance.

    If object attribute is a callable, to get the value it would be
    called without any arguments.
    """

    def get_value(self, obj):
        attr = getattr(obj, self.attr_name)
        if callable(attr):
            return attr()
        return attr


class OrderableListViewMixin(object):
    ordering = None
    ordering_kwarg = "_orderby"

    def get_ordering(self):
        """Return the field or fields to use for ordering the queryset."""
        ordering = []

        # url query parameter
        if self.ordering_kwarg in self.request.GET:
            params = self.request.GET[self.ordering_kwarg].split(",")
            for param in params:
                _, prefix, param_name = param.rpartition("-")
                column_def = self.list_columns.get(param_name)
                if column_def:
                    column_ordering = column_def.orderby()
                    if column_ordering:
                        if hasattr(column_ordering, "as_sql"):
                            ordering.append(
                                column_ordering.desc()
                                if prefix == "-"
                                else column_ordering.asc()
                            )
                        elif column_ordering.startswith("-") and prefix == "-":
                            ordering.append(column_ordering[1:])
                        else:
                            ordering.append(prefix + column_ordering)
        else:
            # default view ordering
            if isinstance(self.ordering, (list, tuple)):
                ordering.extend(self.ordering)
            elif isinstance(self.ordering, str):
                ordering.append(self.ordering)

            # default queryset order
            if self.queryset is not None:
                ordering.extend(self.queryset.query.order_by)

        return ordering

    @cached_property
    def columns_order(self):
        """Return list of columns used to order the queryset."""
        ordering = {}

        # ordered by the url query
        if self.ordering_kwarg in self.request.GET:
            params = self.request.GET[self.ordering_kwarg].split(",")
            for param in params:
                _, param_prefix, param_name = param.rpartition("-")
                column_def = self.list_columns.get(param_name)
                if column_def:
                    column_ordering = column_def.orderby()
                    if column_ordering is not None and isinstance(column_ordering, str):
                        # TODO support custom OrderBy expressions
                        (
                            _,
                            column_order_prefix,
                            column_orderby,
                        ) = column_ordering.rpartition("-")
                        ordering[column_def] = (
                            "asc" if column_order_prefix == param_prefix else "desc"
                        )
        else:
            # ordered by explicit self.ordering definition or by queryset.order_by
            raw_ordering = []
            if isinstance(self.ordering, (list, tuple)):
                raw_ordering.extend(self.ordering)
            elif isinstance(self.ordering, str):
                raw_ordering.append(self.ordering)
            if self.queryset is not None:
                raw_ordering.extend(self.queryset.query.order_by)

            for param in raw_ordering:
                _, param_prefix, param_name = param.rpartition("-")
                for column_def in self.list_columns.values():
                    if column_def in ordering:  # column order already found
                        continue
                    column_ordering = column_def.orderby()
                    if column_ordering is not None and isinstance(column_ordering, str):
                        # TODO support custom OrderBy expressions
                        (
                            _,
                            column_order_prefix,
                            column_orderby,
                        ) = column_ordering.rpartition("-")
                        if param_name == column_orderby:
                            ordering[column_def] = (
                                "asc" if column_order_prefix == param_prefix else "desc"
                            )

        return ordering


class BulkActionsMixin(object):
    bulk_actions = None

    def get_bulk_actions(self, *actions):
        if self.viewset is not None and hasattr(self.viewset, "get_list_bulk_actions"):
            actions = self.viewset.get_list_bulk_actions(self.request) + actions
        if self.bulk_actions:
            actions = self.bulk_actions + actions
        return actions


class BaseListModelView(generic.ListView):
    viewset = None
    columns = None
    object_link_columns = None
    paginate_by = 25

    page_actions = None

    empty_value_display = ""

    def has_view_permission(self, user, obj=None):
        if self.viewset is not None:
            return self.viewset.has_view_permission(user, obj=obj)
        else:
            return has_object_perm(
                user, "view", self.model, obj=obj
            ) or has_object_perm(user, "change", self.model, obj=obj)

    def get_columns(self):
        if self.columns is None:
            return ["__str__"]
        return self.columns

    @lru_cache(maxsize=None)
    def get_object_link_columns(self):
        if self.object_link_columns is None:
            return self.get_columns()[0]
        return self.object_link_columns

    def get_column_def(self, attr_name):
        opts = self.model._meta

        # object printable string representation
        if attr_name == "__str__":
            return ObjectAttrColumn(
                self.model, attr_name, opts.verbose_name.capitalize()
            )

        # a method from view or viewset
        data_sources = [self, self.viewset] if self.viewset is not None else [self]
        for data_source in data_sources:
            if hasattr(data_source, attr_name):
                return DataSourceColumn(data_source, attr_name)

        # an object field
        try:
            model_field = opts.get_field(attr_name)
        except FieldDoesNotExist:
            pass
        else:
            return ModelFieldColumn(model_field)

        # a method from object
        if hasattr(self.model, attr_name):
            return ObjectAttrColumn(self.model, attr_name)

        raise ValueError("Can't found datasource for {} column".format(attr_name))

    def get_object_url(self, obj):
        if self.viewset is not None and hasattr(self.viewset, "get_object_url"):
            return self.viewset.get_object_url(self.request, obj)
        else:
            if hasattr(obj, "get_absolute_url") and self.has_view_perm(
                self.request.user, obj
            ):
                return obj.get_absolute_url()

    @cached_property
    def list_columns(self):
        return {
            column_name: self.get_column_def(column_name)
            for column_name in self.get_columns()
        }

    def format_value(self, obj, column, value):
        result = column.format_value(obj, value)
        if column.attr_name in self.get_object_link_columns():
            url = self.get_object_url(obj)
            if url:
                result = format_html('<a href="{}">{}</a>', url, result)
        return result

    def get_page_data(self, page):
        """Formated page data for a table.

        Returned data is a list of list of cell values zipped with column definitions.
        [[(column, value), (column, value), ...], ...]
        """
        for obj in page:
            yield obj, [
                (
                    column_def,
                    self.format_value(obj, column_def, column_def.get_value(obj)),
                )
                for column_def in self.list_columns.values()
            ]

    def get_page_actions(self, *actions):
        if self.viewset is not None and hasattr(self.viewset, "get_list_page_actions"):
            actions = self.viewset.get_list_page_actions(self.request) + actions
        if self.page_actions:
            actions = self.page_actions + actions
        return actions

    @viewprop
    def queryset(self):
        if self.viewset is not None and hasattr(self.viewset, "get_queryset"):
            return self.viewset.get_queryset(self.request)
        return None

    def get_template_names(self):
        """
        Return a list of template names to be used for the view.

        If `self.template_name` undefined, uses::
             [<app_label>/<model_label>_list.html,
              'viewflow/views/list.html']
        """
        if self.template_name is None:
            opts = self.model._meta
            return [
                "{}/{}{}.html".format(
                    opts.app_label, opts.model_name, self.template_name_suffix
                ),
                "viewflow/views/list.html",
            ]
        return [self.template_name]

    def dispatch(self, request, *args, **kwargs):
        if not self.has_view_permission(self.request.user):
            raise PermissionDenied

        return super().dispatch(request, *args, **kwargs)


@method_decorator(login_required, name="dispatch")
class ListModelView(
    BulkActionsMixin,
    FilterableViewMixin,
    OrderableListViewMixin,
    SearchableViewMixin,
    BaseListModelView,
):
    """
    Render some list of objects.
    """



from .actions import BaseBulkActionView, DeleteBulkActionView
from .base import Action, BulkActionForm, FormLayoutMixin
from .create import CreateModelView
from .delete import DeleteModelView
from .detail import DetailModelView
from .list import ListModelView, FilterableViewMixin, OrderableListViewMixin
from .search import SearchableViewMixin
from .update import UpdateModelView


__all__ = (
    "Action",
    "BaseBulkActionView",
    "BulkActionForm",
    "CreateModelView",
    "DeleteBulkActionView",
    "DeleteModelView",
    "DetailModelView",
    "FilterableViewMixin",
    "FormLayoutMixin",
    "ListModelView",
    "OrderableListViewMixin",
    "SearchableViewMixin",
    "UpdateModelView",
)



from django.conf import settings
from django.contrib.auth import get_user_model
from django.db import models
from phonenumber_field.modelfields import PhoneNumberField

User = get_user_model()


class Department(models.Model):
    created = models.DateTimeField('Создано', auto_now_add=True)
    updated = models.DateTimeField('Изменено', auto_now=True)
    name = models.CharField('Название', max_length=64, unique=True)
    comment = models.CharField('Комментарий', max_length=256, blank=True)

    def __str__(self):
        return self.name

    class Meta:
        managed = False
        verbose_name = 'Подразделение'
        verbose_name_plural = 'Подразделения'


class Ext(models.Model):
    created = models.DateTimeField('Создано', auto_now_add=True)
    updated = models.DateTimeField('Изменено', auto_now=True)
    user = models.OneToOneField(settings.AUTH_USER_MODEL, on_delete=models.PROTECT,
                                verbose_name='Пользователь', primary_key=True)
    phonenumber = PhoneNumberField(unique=True, verbose_name='Номер телефона')
    department = models.ForeignKey(Department, on_delete=models.PROTECT, verbose_name='Подразделение')
    ip = models.GenericIPAddressField()
    http_referer = models.URLField('Реф. ссылка', blank=True, max_length=2048)

    def set_department(self, department_id):
        self.department_id = department_id
        super().save()

    def __str__(self):
        return f'{self.user_id}'

    class Meta:
        managed = False
        verbose_name = 'Расширение пользователя'
        verbose_name_plural = 'Расширение пользователей'



from django.utils.timezone import now

from viewflow import this
from viewflow.workflow.models import Process
from ..activation import Activation
from ..base import Node
from ..signals import task_started, flow_finished
from ..status import STATUS
from .start import StartActivation
from . import mixins


def subprocess_finished(activation):
    return all(
        process.status == STATUS.DONE
        for process in Process._default_manager.filter(
            parent_task=activation.task,
        ).exclude(
            status=STATUS.CANCELED,
        )
    )


class StartSubprocess(mixins.NextNodeMixin, Node):
    """Start subprocess handle"""

    task_type = "START"

    shape = {
        "width": 50,
        "height": 50,
        "svg": """
            <circle class="event" cx="25" cy="25" r="25"/>
        """,
    }

    bpmn_element = "startEvent"

    activation_class = StartActivation

    def __init__(self, func=None, **kwargs):
        super().__init__(**kwargs)
        self._func = func

    def run(self, parent_task, **kwargs):
        func = this.resolve(self.flow_class.instance, self._func)

        def wrapper(**kwargs):
            activation = self.activation_class.create(self, None, parent_task.token)
            activation.process.parent_task = parent_task
            if func:
                func(activation, **kwargs)
            activation.execute()
            return activation

        return wrapper(**kwargs)


class SubprocessActivation(mixins.NextNodeActivationMixin, Activation):
    @Activation.status.transition(source=STATUS.NEW, target=STATUS.STARTED)
    def activate(self):
        with self.exception_guard():
            self.task.started = now()
            self.task.save()
            self.flow_task.start_subprocess_task.run(self.task)
            task_started.send(
                sender=self.flow_class, process=self.process, task=self.task
            )

    @Activation.status.transition(
        source=STATUS.STARTED, target=STATUS.DONE, conditions=[subprocess_finished]
    )
    def complete(self):
        """Complete task and create next."""
        super().complete.original()

    @Activation.status.transition(source=STATUS.STARTED)
    def execute(self):
        self.complete()
        self.activate_next()

    @Activation.status.transition(source=STATUS.ERROR, target=STATUS.DONE)
    def retry(self):
        """Retry the next node calculation and activation."""
        self.activate.original()


class Subprocess(mixins.NextNodeMixin, Node):
    task_type = "SUBPROCESS"
    activation_class = SubprocessActivation

    shape = {
        "width": 150,
        "height": 100,
        "text-align": "middle",
        "svg": """
            <rect class="task" width="150" height="100" rx="5" ry="5" style="stroke-width:5"/>
        """,
    }

    def __init__(self, start_subprocess_task, **kwargs):
        self.start_subprocess_task = start_subprocess_task
        super().__init__(**kwargs)

    def _ready(self):
        flow_finished.connect(
            self.on_flow_finished, sender=self.start_subprocess_task.flow_class
        )

    def on_flow_finished(self, **signal_kwargs):
        process = signal_kwargs["process"]
        if process.parent_task and process.parent_task.flow_task == self:
            with process.parent_task.activation() as activation:
                if activation.complete.can_proceed():
                    activation.execute()


class NSubprocessActivation(SubprocessActivation):
    @Activation.status.transition(source=STATUS.NEW, target=STATUS.STARTED)
    def activate(self):
        with self.exception_guard():
            self.task.started = now()
            self.task.save()

            for item in self.flow_task.subitem_source(self.process):
                self.flow_task.start_subprocess_task.run(self.task, item=item)

            task_started.send(
                sender=self.flow_class, process=self.process, task=self.task
            )


class NSubprocess(Subprocess):
    task_type = "SUBPROCESS"
    activation_class = NSubprocessActivation

    shape = {
        "width": 150,
        "height": 100,
        "text-align": "middle",
        "svg": """
            <rect class="task" width="150" height="100" rx="5" ry="5" style="stroke-width:5"/>
        """,
    }

    def __init__(self, start_subprocess_task, subitem_source, **kwargs):
        self.start_subprocess_task = start_subprocess_task
        self.subitem_source = subitem_source
        super().__init__(start_subprocess_task, **kwargs)



from django.contrib import admin
from django.contrib.admin.widgets import AdminTextInputWidget
from django.db import models as django_models

from .models import Process, Task


class TaskInline(admin.TabularInline):
    """Task inline."""

    model = Task
    fields = ["flow_task", "flow_task_type", "status", "token", "owner"]
    readonly_fields = ["flow_task", "flow_task_type", "status", "token"]
    formfield_overrides = {
        django_models.ForeignKey: {'widget': AdminTextInputWidget},
    }

    def has_add_permission(self, request, obj=None):
        """Disable manually task creation."""
        return False

    def has_delete_permission(self, request, obj=None):
        """Disable task deletion in the process inline."""
        return False


@admin.register(Process)
class ProcessAdmin(admin.ModelAdmin):
    """List all of viewflow process."""

    icon = '<i class="material-icons">assignment</i>'

    actions = None
    date_hierarchy = "created"
    list_display = ["pk", "created", "flow_class", "status", "brief"]
    list_display_links = ["pk", "created", "flow_class"]
    list_filter = ["status", "flow_class"]
    readonly_fields = ["flow_class", "status", "finished", "parent_task"]
    inlines = [TaskInline]
    formfield_overrides = {
        django_models.ForeignKey: {'widget': AdminTextInputWidget},
    }

    def has_add_permission(self, request, obj=None):
        """Disable manually process creation."""
        return False


@admin.register(Task)
class TaskAdmin(admin.ModelAdmin):
    """List all of viewflow tasks."""

    icon = '<i class="material-icons">assignment_turned_in</i>'

    actions = None
    date_hierarchy = "created"
    list_display = [
        "pk",
        "created",
        "process",
        "status",
        "owner",
        "owner_permission",
        "token",
        "started",
        "finished",
    ]
    list_display_links = ["pk", "created", "process"]
    list_filter = ["status", "process__flow_class"]
    readonly_fields = [
        "process",
        "status",
        "flow_task",
        "started",
        "finished",
        "previous",
        "token",
    ]
    formfield_overrides = {
        django_models.ForeignKey: {'widget': AdminTextInputWidget},
    }

    def has_add_permission(self, request, obj=None):
        """Disable manually task creation."""
        return False



from django.urls import path
from viewflow import viewprop
from viewflow.urls import ViewsetMeta
from . import utils


class NodeDetailMixin(metaclass=ViewsetMeta):
    """Task detail view."""

    index_view_class = None

    @viewprop
    def index_view(self):
        """View for a task detail."""
        if self.index_view_class:
            return self.index_view_class.as_view()

    @property
    def index_path(self):
        if self.index_view:
            return path(
                f"<int:process_pk>/{self.name}/<int:task_pk>/",
                utils.wrap_task_view(self, self.index_view),
                name="index",
            )

    detail_view_class = None

    @viewprop
    def detail_view(self):
        """View for a task detail."""
        if self.detail_view_class:
            return self.detail_view_class.as_view()

    @property
    def detail_path(self):
        if self.detail_view:
            return path(
                f"<int:process_pk>/{self.name}/<int:task_pk>/detail/",
                utils.wrap_task_view(self, self.detail_view, permission=self.can_view),
                name="detail",
            )

    def can_view(self, user, task):
        """Check if user has a view task detail permission."""
        return self.flow_class.instance.has_view_permission(user, task)


class NodeExecuteMixin(metaclass=ViewsetMeta):
    """Re-execute a gate manually."""

    execute_view_class = None

    @viewprop
    def execute_view(self):
        """View for the admin to re-execute a gate."""
        if self.execute_view_class:
            return self.execute_view_class.as_view()

    @property
    def execute_path(self):
        if self.execute_view:
            return path(
                "<int:process_pk>/{}/<int:task_pk>/execute/".format(self.name),
                utils.wrap_task_view(self, self.execute_view),
                name="execute",
            )


class NodeUndoMixin(metaclass=ViewsetMeta):
    """Allow to undo a completed task."""

    undo_view_class = None

    @viewprop
    def undo_view(self):
        """View for the admin to undo a task."""
        if self.undo_view_class:
            return self.undo_view_class.as_view()

    @property
    def undo_path(self):
        if self.undo_view:
            return path(
                f"<int:process_pk>/{self.name}/<int:task_pk>/undo/",
                utils.wrap_task_view(self, self.undo_view, permission=self.can_undo),
                name="undo",
            )

    def can_undo(self, user, task):
        return self.flow_class.instance.has_manage_permission(user)


class NodeCancelMixin(metaclass=ViewsetMeta):
    """Cancel a task action."""

    cancel_view_class = None

    @viewprop
    def cancel_view(self):
        """View for the admin to cancel a task."""
        if self.cancel_view_class:
            return self.cancel_view_class.as_view()

    @property
    def cancel_path(self):
        if self.cancel_view:
            return path(
                f"<int:process_pk>/{self.name}/<int:task_pk>/cancel/",
                utils.wrap_task_view(self, self.cancel_view, permission=self.can_cancel),
                name="cancel",
            )

    def can_cancel(self, user, task):
        return self.flow_class.instance.has_manage_permission(user)


class NodeReviveMixin(metaclass=ViewsetMeta):
    """Review a canceled task"""

    revive_view_class = None

    @viewprop
    def revive_view(self):
        """View for the admin to cancel a task."""
        if self.revive_view_class:
            return self.revive_view_class.as_view()

    @property
    def revive_path(self):
        if self.revive_view:
            return path(
                f"<int:process_pk>/{self.name}/<int:task_pk>/revive/",
                utils.wrap_task_view(self, self.revive_view, permission=self.can_revive),
                name="revive",
            )

    def can_revive(self, user, task):
        return self.flow_class.instance.has_manage_permission(user)





from django.apps import AppConfig

# Copyright (c) 2017-2020, Mikhail Podgurskiy
# All Rights Reserved.

# This work is dual-licensed under AGPL defined in file 'LICENSE' with
# LICENSE_EXCEPTION and the Commercial licence defined in file 'COMM_LICENSE',
# which is part of this source code package.


class ViewflowConfig(AppConfig):
    """Default application config."""

    name = 'viewflow'
    label = 'viewflow_base'  # allow to user 'viewflow' label for 'viewflow.workflow'



import logging

from pythonjsonlogger import jsonlogger


class YandexLoggingFormatter(jsonlogger.JsonFormatter):
    def add_fields(self, log_record, record, message_dict):
        super(YandexLoggingFormatter, self).add_fields(log_record, record, message_dict)
        log_record['logger'] = record.name
        log_record['level'] = str.replace(str.replace(record.levelname, "WARNING", "WARN"), "CRITICAL", "FATAL")


class YandexStreamHandler(logging.StreamHandler):
    def format(self, record):
        self.formatter = YandexLoggingFormatter('%(message)s %(level)s %(logger)s')
        return super().format(record)



"""Forward-reference for class-body declarations"""

# Copyright (c) 2017-2020, Mikhail Podgurskiy
# All Rights Reserved.

# This work is dual-licensed under AGPL defined in file 'LICENSE' with
# LICENSE_EXCEPTION and the Commercial license defined in file 'COMM_LICENSE',
# which is part of this source code package.
from typing import Any


class ThisMethod(object):
    """Reference to a method"""
    def __init__(self, propname, methodname):
        self._propname = propname
        self._methodname = methodname

    def resolve(self, instance):
        # TODO meaningfull exception
        prop = getattr(instance, self._propname)
        method = getattr(instance, f'_this_{self._methodname}')
        return method(prop)


class ThisObject(object):
    """Helper for forward references"""

    def __init__(self, name):  # noqa D102
        self.name = name

    def resolve(self, instance: object) -> Any:
        # TODO meaningfull exception
        return getattr(instance, self.name)

    # def __copy__(self):
    #     return super().__copy__()

    # def __deepcopy__(self, memo):
    #     return super().__deepcopy__(memo)

    def __getattr__(self, name):
        if name.startswith('__'):
            super().__getattr__(name)
        return ThisMethod(self.name, name)


class This(object):
    """Helper for building forward references.

    The rationale is ability to specify references to the class
    attributes and methods before they are declared.

    `this` is like a `self` but for the class body.
    """
    def resolve(self, instance, this_ref):
        if isinstance(this_ref, (ThisObject, ThisMethod)):
            return this_ref.resolve(instance)
        else:
            return this_ref

    def __getattr__(self, name):
        return ThisObject(name)


this = This()



from typing import Callable, List, Optional, Union

from django.apps import apps
from django.contrib import admin
from django.http import HttpResponse
from django.shortcuts import redirect
from django.urls import reverse
from django.utils.safestring import mark_safe
from django.utils.text import capfirst


class InlineActionException(Exception):
    pass


class ActionNotCallable(InlineActionException):
    def __init__(self, model_admin, action, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model_admin = model_admin
        self.action = action


class BaseInlineActionsMixin:
    INLINE_MODEL_ADMIN = 'inline'
    MODEL_ADMIN = 'admin'

    inline_actions: Optional[List[Union[str, Callable]]] = []

    def get_inline_actions(self, request, obj=None):
        """
        Returns a list of all actions for this Admin.
        """
        # If self.actions is explicitly set to None that means that we don't
        # want *any* actions enabled on this page.
        if self.inline_actions is None:
            return []

        actions = []

        # Gather actions from the inline admin and all parent classes,
        # starting with self and working back up.
        for klass in self.__class__.mro()[::-1]:
            class_actions = getattr(klass, 'inline_actions', [])
            # Avoid trying to iterate over None
            if not class_actions:
                continue

            for action in class_actions:
                if action not in actions:
                    actions.append(action)

        return actions

    def get_readonly_fields(self, request, obj=None):
        fields = super().get_readonly_fields(request, obj)
        fields = list(fields)

        if 'render_inline_actions' not in fields:
            fields.append('render_inline_actions')
        return fields

    def _get_admin_type(self, model_admin=None):
        """
        Returns wether this is an InlineAdmin or not.
        """
        model_admin = model_admin or self

        if isinstance(model_admin, admin.options.InlineModelAdmin):
            return self.INLINE_MODEL_ADMIN
        return self.MODEL_ADMIN

    def render_inline_actions(self, obj=None):  # NOQA: C901
        """
        Renders all defined inline actions as html.
        """
        if not (obj and obj.pk):
            return ''

        buttons = []
        for action_name in self.get_inline_actions(self._request, obj):
            action_func = getattr(self, action_name, None)
            if not action_func:
                raise RuntimeError("Could not find action `{}`".format(action_name))

            # Add per-object label support
            action_name = action_func.__name__
            label_handler = getattr(self, 'get_{}_label'.format(action_name), None)
            if callable(label_handler):
                description = label_handler(obj=obj)
            else:
                try:
                    description = action_func.short_description
                except AttributeError:
                    description = capfirst(action_name.replace('_', ' '))

            # Add per-object css classes support
            css_handler = getattr(self, 'get_{}_css'.format(action_name), None)
            if callable(css_handler):
                css_classes = css_handler(obj=obj)
            else:
                try:
                    css_classes = action_func.css_classes
                except AttributeError:
                    css_classes = ''

            # If the form is submitted, we have no information about the
            # requested action.
            # Hence we need all data to be encoded using the action name.
            action_data = [
                # required to distinguish between multiple inlines for the same model
                self.__class__.__name__.lower(),
                self._get_admin_type(),
                action_name,
                obj._meta.app_label,
                obj._meta.model_name,
                str(obj.pk),
            ]
            buttons.append(
                '<input type="submit" name="{}" value="{}" class="{}">'.format(
                    '_action__{}'.format('__'.join(action_data)),
                    description,
                    css_classes,
                )
            )
        return mark_safe(
            '<div class="submit_row inline_actions">{}</div>'.format(''.join(buttons))
        )

    render_inline_actions.short_description = "Действия"  # type: ignore
    render_inline_actions.allow_tags = True  # type: ignore


class InlineActionsMixin(BaseInlineActionsMixin):
    def render_inline_actions(self, obj=None):
        html = super().render_inline_actions(obj=obj)
        # we have to add <p> tags as a workaround for invalid html
        return mark_safe('</p>{}<p>'.format(html))

    render_inline_actions.short_description = "Действия"  # type: ignore
    render_inline_actions.allow_tags = True  # type: ignore

    def get_fields(self, request, obj=None):
        # store `request` for `get_inline_actions`
        self._request = request

        fields = super().get_fields(request, obj)
        if self.inline_actions is not None:  # is it explicitly disabled?
            fields = list(fields)
            if 'render_inline_actions' not in fields:
                fields.append('render_inline_actions')
        return fields


class InlineActionsModelAdminMixin(BaseInlineActionsMixin):
    class Media:
        css = {"all": ("inline_actions/css/inline_actions.css",)}

    def get_list_display(self, request):
        # store `request` for `get_inline_actions`
        self._request = request

        fields = super().get_list_display(request)
        if self.inline_actions is not None:  # is it explicitly disabled?
            fields = list(fields)
            if 'render_inline_actions' not in fields:
                fields.append('render_inline_actions')
        return fields

    def get_fields(self, request, obj=None):
        # store `request` for `get_inline_actions`
        self._request = request

        fields = super().get_fields(request, obj=obj)
        if not self.fields:
            # django adds all readonly fields by default
            # if `self.fields` is not defined we don't want to include
            # `render_inline_actions
            if 'render_inline_actions' in fields:
                fields.remove('render_inline_actions')
        return fields

    def _execute_action(self, request, model_admin, action, obj, parent_obj=None):
        """
        Tries to execute the requested action and returns a `HttpResponse`.

        raises
            ActionNotCallable - When action is not a function
        """
        # execute action
        func = getattr(model_admin, action, None)
        try:
            response = func(request, obj, parent_obj=parent_obj)
        except TypeError as e:
            raise ActionNotCallable(model_admin, action) from e

        # we should receive an HttpResponse
        if isinstance(response, HttpResponse):
            return response

        # otherwise redirect back
        if parent_obj is None:  # InlineActionsMixin.MODEL_ADMIN:
            # redirect to `changelist`
            url = reverse(
                'admin:{}_{}_changelist'.format(
                    obj._meta.app_label,
                    obj._meta.model_name,
                ),
            )
        else:
            # redirect to `changeform`
            url = reverse(
                'admin:{}_{}_change'.format(
                    parent_obj._meta.app_label,
                    parent_obj._meta.model_name,
                ),
                args=(parent_obj.pk,),
            )

        # readd query string
        query = request.META['QUERY_STRING'] or request.GET.urlencode()
        if query:
            url = '{}?{}'.format(url, query)

        return redirect(url)

    def _handle_action(self, request, object_id=None):
        """
        Resolve and executes the action issued by the current request.
        If no action was triggered, it does nothing.

        Returns `HttpResponse` or `None`
        """
        all_actions = [
            key for key in list(request.POST.keys()) if key.startswith('_action__')
        ]

        if request.method == 'POST' and all_actions:
            assert len(all_actions) == 1
            raw_action_name = all_actions[0].replace('_action__', '', 1)

            # resolve action and target models
            raw_action_parts = raw_action_name.split('__')
            admin_class_name, admin_type = raw_action_parts[:2]
            action, app_label, model_name, object_pk = raw_action_parts[2:]

            model = apps.get_model(app_label=app_label, model_name=model_name)
            parent_obj = self.get_object(request, object_id)

            # find action and execute
            if admin_type == self.MODEL_ADMIN:
                model_admin = self
                obj = model_admin.get_queryset(request).get(pk=object_pk)
                # parent_obj is None because `object_id` is None

            else:
                for inline in self.get_inline_instances(request):
                    inline_class_name = inline.__class__.__name__.lower()
                    matches_inline_class = inline_class_name == admin_class_name
                    matches_model = inline.model == model

                    if not matches_model or not matches_inline_class:
                        continue
                    model_admin = inline
                obj = model_admin.get_queryset(request).get(pk=object_pk)

            if model_admin:
                return self._execute_action(
                    request, model_admin, action, obj, parent_obj
                )
        return None

    def changeform_view(self, request, object_id=None, form_url='', extra_context=None):
        # handle requested action if required
        response = self._handle_action(request, object_id=object_id)
        if response:
            return response

        # continue normally
        return super().changeform_view(request, object_id, form_url, extra_context)

    def changelist_view(self, request, extra_context=None):
        # handle requested action if required
        response = self._handle_action(request)
        if response:
            return response

        # continue normally
        return super().changelist_view(request, extra_context)



from django import template
from django.utils.safestring import mark_safe

register = template.Library()


@register.simple_tag(takes_context=True)
def render_inline_action_fields(context):
    """
    Render hidden fields, which are required for identifying an action.
    """
    request = context.get('request', {'POST': {}})
    all_actions = [
        key for key in list(request.POST.keys()) if key.startswith('_action__')
    ]
    if len(all_actions) == 0:
        raise RuntimeError("No inline action has been triggered.")
    if len(all_actions) > 1:
        raise RuntimeError(
            "Multiple inline actions have been triggered simultaneously."
        )

    action_key = all_actions[0]
    fields = '<input type="hidden" name="{}" value="">'.format(action_key)
    return mark_safe(fields)



from django.apps import AppConfig


class ProfilesConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'expressmoney.apps.ru_profiles.profiles'
    verbose_name = 'Профили'



# Copyright (c) 2017-2020, Mikhail Podgurskiy
# All Rights Reserved.

# This work is dual-licensed under AGPL defined in file 'LICENSE' with
# LICENSE_EXCEPTION and the Commercial license defined in file 'COMM_LICENSE',
# which is part of this source code package.

from django.db import router
from django.db.models.deletion import Collector
from django.contrib import messages
from django.http import HttpResponseRedirect
from django.utils.functional import cached_property
from django.views import generic
from django.views.generic.list import MultipleObjectMixin
from .base import BulkActionForm
from .list import FilterableViewMixin


class BaseBulkActionView(FilterableViewMixin, MultipleObjectMixin, generic.FormView):
    model = None
    viewset = None

    form_class = BulkActionForm
    template_name_suffix = '_action'

    def get_template_names(self):
        opts = self.model._meta
        return [
            '{}/{}{}.html'.format(
                opts.app_label,
                opts.model_name,
                self.template_name_suffix),
            self.template_name,
        ]

    def get_success_url(self):
        if self.viewset and hasattr(self.viewset, 'get_success_url'):
            return self.viewset.get_success_url(self.request)
        return '../'

    def get_form_kwargs(self):
        return {
            'model': self.model,
            **super().get_form_kwargs(),
        }

    def get_queryset(self):
        queryset = super().get_queryset()
        pks = self.request.POST.getlist('pk')
        select_all = self.request.POST.get('select_all')
        if not select_all and pks:
            queryset = queryset.filter(pk__in=pks)
        self.object_list = queryset
        return queryset

    @cached_property
    def objects_count(self):
        if self.request.POST.get('select_all') and not self.filterset.form.has_changed():
            return None
        return self.object_list.count()

    def form_not_confirmed(self, form):
        return self.render_to_response(self.get_context_data(form=form))

    def get(self, request, *args, **kwargs):
        messages.add_message(self.request, messages.SUCCESS, 'No objects selected', fail_silently=True)
        return HttpResponseRedirect(self.get_success_url())

    def post(self, request, *args, **kwargs):
        """
        Handles POST requests, instantiating a form instance with the passed
        POST variables and then checked for validity.
        """
        self.form = self.get_form()
        self.object_list = self.get_queryset()
        if self.form.is_valid():
            if '_confirm' in self.request.POST:
                return self.form_valid(self.form)
            else:
                return self.form_not_confirmed(self.form)
        else:
            return self.form_invalid(self.form)


class DeleteBulkActionView(BaseBulkActionView):
    template_name = 'viewflow/views/delete_action.html'
    template_name_suffix = '_delete_action'

    def get_deleted_objects(self, query):
        collector = Collector(using=router.db_for_write(self.model))
        collector.collect(query)
        return [
            (model_class, objects)
            for model_class, objects in collector.data.items()
        ]

    def get_context_data(self, **kwargs):
        """Extend view context data.
        `{{ deleted_objects }}` - list of related objects to delete
        """
        if self.form.is_valid() and not self.form.cleaned_data.get('select_all'):
            kwargs.setdefault('deleted_objects', self.get_deleted_objects(self.get_queryset()))
        return super(DeleteBulkActionView, self).get_context_data(**kwargs)

    def form_valid(self, form):
        self.get_queryset().delete()
        self.message_user()
        return HttpResponseRedirect(self.get_success_url())

    def message_user(self):
        message = "The objects was deleted successfully"
        messages.add_message(self.request, messages.SUCCESS, message, fail_silently=True)



from django.db import models
from django.db.models import Q, F, Value
from django.db.models.functions import Concat
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.db.models.query import QuerySet
from django.db.models.constants import LOOKUP_SEP
from django.db.models.query import ModelIterable

from .status import STATUS
from .utils import get_next_process_task


def _available_flows(flow_classes, user):
    result = []
    for flow_class in flow_classes:
        if flow_class.instance.has_view_permission(user):
            result.append(flow_class)
    return result


def _get_related_path(model, base_model):
    """Return path suitable for select related for subclass."""
    ancestry = []

    if model._meta.proxy:
        for parent in model._meta.get_parent_list():
            if not parent._meta.proxy:
                model = parent
                break

    parent = model._meta.get_ancestor_link(base_model)

    while parent is not None:
        related = parent.remote_field if hasattr(parent, "remote_field") else parent.rel

        ancestry.insert(0, related.get_accessor_name())
        parent_model = related.model

        parent = parent_model._meta.get_ancestor_link(base_model)

    return LOOKUP_SEP.join(ancestry)


def _get_sub_obj(obj, query):
    rel, _, query = query.partition(LOOKUP_SEP)

    try:
        node = getattr(obj, rel)
    except ObjectDoesNotExist:
        return None

    if query:
        return _get_sub_obj(node, query)
    else:
        return node


def coerce_to_related_instance(instance, target_model):
    """Return subclass of the base object."""
    related = _get_related_path(target_model, instance.__class__)
    if related:
        instance = _get_sub_obj(instance, related)
    if instance and not isinstance(instance, target_model):
        # Coerce proxy classes
        instance.__class__ = target_model
    return instance


class ProcessIterable(ModelIterable):
    def __iter__(self):
        base_iterator = super().__iter__()
        if getattr(self.queryset, "_coerced", False):
            for process in base_iterator:
                if isinstance(process, self.queryset.model):
                    process = coerce_to_related_instance(
                        process, process.flow_class.process_class
                    )
                yield process
        else:
            for process in base_iterator:
                yield process


class ProcessQuerySet(QuerySet):
    """Base manager for the flow Process."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._iterable_class = ProcessIterable

    def coerce_for(self, flow_classes):
        """Return subclass instances of the Task."""
        self._coerced = True

        flow_classes = list(flow_classes)

        related = filter(
            None,
            map(
                lambda flow_class: _get_related_path(
                    flow_class.process_class, self.model
                ),
                flow_classes,
            ),
        )

        return self.filter(flow_class__in=flow_classes).select_related(*related)

    def filter_available(self, flow_classes, user):
        """List of processes available to view for the user."""
        return self.model.objects.coerce_for(_available_flows(flow_classes, user))

    def _chain(self, **kwargs):
        chained = super()._chain(**kwargs)
        if hasattr(self, "_coerced"):
            chained._coerced = self._coerced
        return chained


class TaskIterable(ModelIterable):
    def __iter__(self):
        base_iterator = super().__iter__()
        if getattr(self.queryset, "_coerced", False):
            for task in base_iterator:
                if isinstance(task, self.queryset.model):
                    task = coerce_to_related_instance(
                        task, task.flow_task.flow_class.task_class
                    )
                yield task
        else:
            for task in base_iterator:
                yield task


class TaskQuerySet(QuerySet):
    """Base manager for the Task."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._iterable_class = TaskIterable

    def coerce_for(self, flow_classes):
        """Return subclass instances of the Task."""
        self._coerced = True
        flow_classes = list(flow_classes)

        related = filter(
            None,
            map(
                lambda flow_class: _get_related_path(flow_class.task_class, self.model),
                flow_classes,
            ),
        )

        return self.filter(process__flow_class__in=flow_classes).select_related(
            "process", *related
        )

    def user_queue(self, user, flow_class=None):
        """List of tasks of the flow_class permitted for user."""
        queryset = self.filter(flow_task_type="HUMAN")

        if flow_class is not None:
            queryset = queryset.filter(process__flow_class=flow_class)

        if not user.is_superuser:
            user_id = user.id
            has_permission = Q(owner_permission__isnull=True) | Q(owner=user_id)

            if "guardian" in settings.INSTALLED_APPS:
                from guardian.models import UserObjectPermission, GroupObjectPermission

                queryset = queryset.annotate(
                    owner_permission_obj_check=Concat(
                        F("owner_permission"),
                        Value("_"),
                        F("owner_permission_obj_pk"),
                        output_field=models.CharField(),
                    )
                )

                all_user_perms = user.get_all_permissions()
                if hasattr(user, "_viewflow_per_object_perm_cache"):
                    per_object_perms = user._viewflow_per_object_perm_cache
                else:
                    per_object_perms = {
                        "{}.{}_{}".format(
                            userperm.content_type.app_label,
                            userperm.permission.codename,
                            userperm.object_pk,
                        )
                        for userperm in UserObjectPermission.objects.filter(user=user_id)
                    }

                    per_object_perms = per_object_perms | {
                        "{}.{}_{}".format(
                            groupperm.content_type.app_label,
                            groupperm.permission.codename,
                            groupperm.object_pk,
                        )
                        for groupperm in GroupObjectPermission.objects.filter(
                            group__in=user.groups.all()
                        )
                    }
                    user._viewflow_per_object_perm_cache = per_object_perms

                has_permission = (
                    has_permission
                    | Q(owner_permission__in=all_user_perms)
                    | Q(owner_permission_obj_check__in=per_object_perms)
                )
            else:
                has_permission = (
                    Q(owner_permission__in=user.get_all_permissions()) | has_permission
                )

            queryset = queryset.filter(has_permission)

        return queryset

    def user_archive(self, user, flow_class=None):
        """List of tasks of the flow_class completed by the user."""
        queryset = self.filter(flow_task_type="HUMAN")

        if flow_class is not None:
            queryset = queryset.filter(process__flow_class=flow_class)

        return queryset.filter(owner=user.pk, finished__isnull=False)

    def filter_available(self, flow_classes, user):
        """List of tasks available to view for the user."""
        return self.model.objects.coerce_for(_available_flows(flow_classes, user))

    def inbox(self, flow_classes, user):
        """List of tasks assigned to the user."""
        return self.filter_available(flow_classes, user).filter(
            owner=user, status=STATUS.ASSIGNED
        )

    def queue(self, flow_classes, user):
        """List of tasks permitted to assign for the user."""
        return (
            self.filter_available(flow_classes, user)
            .user_queue(user)
            .filter(status=STATUS.NEW)
        )

    def archive(self, flow_classes, user):
        """List of tasks finished by the user."""
        return self.filter_available(flow_classes, user).filter(
            owner=user, finished__isnull=False
        )

    def next_user_task(self, process, user):
        """
        Lookup for the next task for a user execution.

        Prefer assigned tasks first, if not, return first task from user queue
        """

        # task inside a same process
        task = get_next_process_task(self, process, user)

        # task inside subprocess
        if task is None:
            subprocess_task = self.filter(process__parent_task__process=process).first()
            if subprocess_task:
                task = get_next_process_task(self, subprocess_task.process, user)

        # task inside parent process
        if task is None and process.parent_task_id:
            task = get_next_process_task(self, process.parent_task.process, user)

        # task inside other subprocesses of parent task
        if task is None and process.parent_task_id:
            processes = process.__class__._default_manager.filter(
                parent_task__process=process.parent_task.process
            ).exclude(pk=process.pk)
            for sub_process in processes:
                task = get_next_process_task(self, sub_process, user)
                if task:
                    break

        return task

    def _chain(self, **kwargs):
        chained = super()._chain(**kwargs)
        if hasattr(self, "_coerced"):
            chained._coerced = self._coerced
        return chained



from django.http import HttpResponseRedirect
from django.views import generic
from django.utils.translation import gettext_lazy as _
from viewflow.views import FormLayoutMixin

from . import mixins


class CreateProcessView(
    FormLayoutMixin,
    mixins.SuccessMessageMixin,
    mixins.TaskSuccessUrlMixin,
    mixins.TaskViewTemplateNames,
    generic.UpdateView,
):
    """Default view to start a process"""

    success_message = _("Process {process} has been started.")
    template_filename = "start.html"

    def get_object(self):
        """Return the process for the task activation."""
        return self.request.activation.process

    def form_valid(self, form):
        """If the form is valid, save the associated model and finish the task."""
        self.object = form.save()
        self.request.activation.execute()
        return HttpResponseRedirect(self.get_success_url())


class CreateArtifactView(
    FormLayoutMixin,
    mixins.SuccessMessageMixin,
    mixins.TaskSuccessUrlMixin,
    mixins.TaskViewTemplateNames,
    generic.CreateView,
):
    template_filename = "start.html"

    def form_valid(self, form):
        self.object = form.save()
        self.request.activation.process.artifact = self.object
        self.request.activation.execute()
        return HttpResponseRedirect(self.get_success_url())
